from math import pi
from collections import namedtuple

# https://en.wikipedia.org/wiki/Earth_radius#Mean_radius
EARTH_MEAN_RADIUS = 6371008.8
EARTH_MEAN_DIAMETER = 2 * EARTH_MEAN_RADIUS

# https://en.wikipedia.org/wiki/Earth_radius#Equatorial_radius
EARTH_EQUATORIAL_RADIUS = 6378137.0
EARTH_EQUATORIAL_METERS_PER_DEGREE = pi * EARTH_EQUATORIAL_RADIUS / 180 # 111319.49079327358
I_EARTH_EQUATORIAL_METERS_PER_DEGREE = 1 / EARTH_EQUATORIAL_METERS_PER_DEGREE

HALF_PI = pi / 2.0
QUARTER_PI = pi / 4.0

# https://en.wikipedia.org/wiki/Geodetic_datum
Datum = namedtuple('Datum', ['a', 'b', 'e', 'f', 'w'])

# https://epsg.io/7030-ellipsoid
WGS84 = Datum(
    # equatorial radius (semi-major axis)
    a=6378137.0,
    # polar radius (semi-minor axis)
    b=6356752.314245179, # b = a * (1 - f)
# derived ellipsoid parameters
    # eccentricity
    e = 0.08181919084262149, # e = (2*f - f**2)**0.5
    # flattening
    f=0.0033528106647474805, # 1/298.257223563
    # rotation speed
    w=7292115e-11, # rad/sec;
)



from math import (
    degrees, radians,
    sin, cos, asin, tan, atan, atan2, pi,
    sqrt, exp, log, fabs
)
from geo.constants import WGS84, QUARTER_PI, HALF_PI


CONVERGENCE_THRESHOLD = 1e-12
MAX_ITERATIONS = 15


def _py_distance(point1, point2, ellipsoid=WGS84):
    '''
        Calculating distance with using vincenty's formula
        https://en.wikipedia.org/wiki/Vincenty's_formulae
    '''
    lon1, lat1 = (radians(coord) for coord in point1)
    lon2, lat2 = (radians(coord) for coord in point2)

    U1 = atan((1 - ellipsoid.f) * tan(lat1))
    U2 = atan((1 - ellipsoid.f) * tan(lat2))
    L = lon2 - lon1
    Lambda = L

    sinU1 = sin(U1)
    cosU1 = cos(U1)
    sinU2 = sin(U2)
    cosU2 = cos(U2)

    for _ in range(MAX_ITERATIONS):
        sinLambda = sin(Lambda)
        cosLambda = cos(Lambda)
        sinSigma = sqrt(
            (cosU2 * sinLambda) ** 2 +
            (cosU1 * sinU2 - sinU1 * cosU2 * cosLambda) ** 2)
        # coincident points
        if sinSigma == 0:
            return 0.0

        cosSigma = sinU1 * sinU2 + cosU1 * cosU2 * cosLambda
        sigma = atan2(sinSigma, cosSigma)
        sinAlpha = cosU1 * cosU2 * sinLambda / sinSigma
        cosSqAlpha = 1 - sinAlpha ** 2
        try:
            cos2SigmaM = cosSigma - 2 * sinU1 * sinU2 / cosSqAlpha
        except ZeroDivisionError:
            cos2SigmaM = 0

        C = (ellipsoid.f / 16) * cosSqAlpha * (
            4 + ellipsoid.f * (4 - 3 * cosSqAlpha)
        )
        LambdaPrev = Lambda
        Lambda = (
            L + (1 - C) * ellipsoid.f * sinAlpha * (
                sigma + C * sinSigma * (
                    cos2SigmaM + C * cosSigma * (
                        -1 + 2 * cos2SigmaM ** 2
                    )
                )
            )
        )

        if abs(Lambda - LambdaPrev) < CONVERGENCE_THRESHOLD:
            break
    else:
        # failure to converge
        return None

    uSq = cosSqAlpha * (ellipsoid.a ** 2 - ellipsoid.b ** 2) / (ellipsoid.b ** 2)
    A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)))
    B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)))
    deltaSigma = B * sinSigma * (cos2SigmaM + B / 4 * (cosSigma *
                 (-1 + 2 * cos2SigmaM ** 2) - B / 6 * cos2SigmaM *
                 (-3 + 4 * sinSigma ** 2) * (-3 + 4 * cos2SigmaM ** 2)))
    s = ellipsoid.b * A * (sigma - deltaSigma)
    return s


def _py_from4326_to3395(point, ellipsoid=WGS84):
    '''
        https://en.wikipedia.org/wiki/Mercator_projection#Generalization_to_the_ellipsoid
        http://epsg.io/3395

        Ellipsoidal Mercator:
            E = a*(λ - λo)
            N = a*ln(tan(π/4+φ/2)*((1-e*sin(φ))/(1+e*sin(φ)))**e/2)
    '''
    lon, lat = (radians(coord) for coord in point)
    e = ellipsoid.e
    a = ellipsoid.a
    e_sin_lat = e*sin(lat)
    multiplier1 = tan(QUARTER_PI + lat / 2.0)
    multiplier2 = pow((1-e_sin_lat)/(1+e_sin_lat),e/2)
    E = a * lon
    N = a*log(multiplier1*multiplier2)
    return (E, N)


def _py_from3395_to4326(point, ellipsoid=WGS84):
    '''
        Reverse Ellipsoidal Mercator:
            λ = E/a + λo
            φ = π/2 + 2*arctan(exp(-N/a)*((1-e*sin(φ))/(1+e*sin(φ))**e/2))
    '''
    E, N = point
    e = ellipsoid.e
    a = ellipsoid.a
    half_e = e * 0.5

    m1 = exp(-N / a)
    old_phi = HALF_PI - 2.0 * atan(m1)

    for _ in range(MAX_ITERATIONS):
        e_sin_phi = e*sin(old_phi)
        m2 = ((1 - e_sin_phi) / (1 + e_sin_phi))**half_e
        phi = HALF_PI  - 2.0 * atan(m1 * m2)
        if abs(old_phi - phi) <= CONVERGENCE_THRESHOLD:
            break
        old_phi = phi

    lon = degrees(E / a)
    lat = degrees(phi)
    return (lon, lat)


try:
    from ._ellipsoid import (
        _distance,
        _from4326_to3395,
        _from3395_to4326
    )
    distance = _distance
    from4326_to3395 = _from4326_to3395
    from3395_to4326 = _from3395_to4326
except ImportError:  # pragma: no cover
    distance = _py_distance
    from4326_to3395 = _py_from4326_to3395
    from3395_to4326 = _py_from3395_to4326



import os
import sys
from setuptools import setup, Extension

try:
    from Cython.Build import cythonize
    ext = '.pyx'
except ImportError:
    def cythonize(extensions): return extensions
    ext = '.c'

MIN_PYTHON = (3, 0)
if sys.version_info < MIN_PYTHON:
    sys.stderr.write("Python {}.{} or later is required\n".format(*MIN_PYTHON))
    sys.exit(1)

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

extensions = cythonize([
    Extension('geo._sphere', ['geo/_sphere' + ext]),
    Extension('geo._ellipsoid', ['geo/_ellipsoid' + ext]),
])


setup(
    name='geo-py',
    version='0.3',
    author='Alexander Verbitsky',
    author_email='habibutsu@gmail.com',
    maintainer='Alexander Verbitsky',
    maintainer_email='habibutsu@gmail.com',
    description='Set of algorithms and structures related to geodesy and geospatial data',
    long_description=read('README.rst'),
    keywords='geodesy, haversine distance, great circle distance, vincenty\'s formula',
    url='https://github.com/gojuno/geo-py',
    packages=['geo'],
    ext_modules = extensions,
    test_suite='test',
    license='BSD',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'License :: OSI Approved :: BSD License',
    ],
)



import unittest
import time
import logging
import math

import geo.sphere as sphere
import geo.ellipsoid as ellipsoid
import geo._sphere as csphere

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()

p_minsk = (27.561831, 53.902257)
p_moscow = (37.620393, 55.75396)


class TestSphere(unittest.TestCase):

    def test_distance(self):
        assert math.isclose(
            sphere._py_approximate_distance(p_minsk, p_moscow),
            676371.322420,
            rel_tol=1e-06)
        assert math.isclose(
            sphere._approximate_distance(p_minsk, p_moscow),
            676371.322420,
            rel_tol=1e-06)

        assert math.isclose(
            sphere._py_haversine_distance(p_minsk, p_moscow),
            675656.299481,
            rel_tol=1e-06)
        assert math.isclose(
            sphere._haversine_distance(p_minsk, p_moscow),
            675656.299481,
            rel_tol=1e-06)

        assert math.isclose(
            sphere._py_distance(p_minsk, p_moscow),
            675656.299481,
            rel_tol=1e-06)
        assert math.isclose(
            sphere._distance(p_minsk, p_moscow),
            675656.299481,
            rel_tol=1e-06)

    def test_projection(self):
        x, y = sphere._py_from4326_to3857(p_minsk)
        assert math.isclose(x, 3068168.9922502628, rel_tol=1e-06)
        assert math.isclose(y, 7151666.629430503, rel_tol=1e-06)
        x, y = sphere._from4326_to3857(p_minsk)
        assert math.isclose(x, 3068168.9922502628, rel_tol=1e-06)
        assert math.isclose(y, 7151666.629430503, rel_tol=1e-06)

        lon, lat = sphere._py_from3857_to4326(
            sphere._py_from4326_to3857(p_minsk))
        assert math.isclose(lon, p_minsk[0], rel_tol=1e-06)
        assert math.isclose(lat, p_minsk[1], rel_tol=1e-06)

        lon, lat = sphere._from3857_to4326(
            sphere._from4326_to3857(p_minsk))
        assert math.isclose(lon, p_minsk[0], rel_tol=1e-06)
        assert math.isclose(lat, p_minsk[1], rel_tol=1e-06)


class TestEllipsoid(unittest.TestCase):

    def test_distance(self):
        assert math.isclose(
            ellipsoid._py_distance(p_minsk, p_moscow),
            677789.531233,
            rel_tol=1e-06)
        assert math.isclose(
            ellipsoid._distance(p_minsk, p_moscow),
            677789.531233,
            rel_tol=1e-06)

    def test_projection(self):
        assert (
            ellipsoid._py_from4326_to3395(p_minsk) ==
            (3068168.9922502623, 7117115.955611216)
        )
        rp_minsk = ellipsoid._py_from3395_to4326(
                ellipsoid._py_from4326_to3395(p_minsk))

        assert math.isclose(rp_minsk[0], p_minsk[0], rel_tol=1e-06)
        assert math.isclose(rp_minsk[1], p_minsk[1], rel_tol=1e-06)

        assert (
            ellipsoid._from4326_to3395(p_minsk) ==
            (3068168.9922502623, 7117115.955611216)
        )

        rp_minsk = ellipsoid._from3395_to4326(
                ellipsoid._from4326_to3395(p_minsk))

        assert math.isclose(rp_minsk[0], p_minsk[0], rel_tol=1e-06)
        assert math.isclose(rp_minsk[1], p_minsk[1], rel_tol=1e-06)



    # -*- coding: utf-8 -*-
from math import (
    degrees, radians,
    sin, cos, asin, tan, atan, atan2, pi,
    sqrt, exp, log, fabs
)

from geo.constants import (
    EARTH_MEAN_RADIUS,
    EARTH_MEAN_DIAMETER,
    EARTH_EQUATORIAL_RADIUS,
    EARTH_EQUATORIAL_METERS_PER_DEGREE,
    I_EARTH_EQUATORIAL_METERS_PER_DEGREE,
    HALF_PI,
    QUARTER_PI,
)

def _py_approximate_distance(point1, point2):
    '''
        Approximate calculation distance
        (expanding the trigonometric functions around the midpoint)
    '''

    lon1, lat1 = (radians(coord) for coord in point1)
    lon2, lat2 = (radians(coord) for coord in point2)
    cos_lat = cos((lat1+lat2)/2.0)
    dx = (lat2 - lat1)
    dy = (cos_lat*(lon2 - lon1))
    return EARTH_MEAN_RADIUS*sqrt(dx**2 + dy**2)


def _py_haversine_distance(point1, point2):
    '''
        Calculating haversine distance between two points
        (see https://en.wikipedia.org/wiki/Haversine_formula,
            https://www.math.ksu.edu/~dbski/writings/haversine.pdf)

        Is numerically better-conditioned for small distances
    '''
    lon1, lat1 = (radians(coord) for coord in point1[:2])
    lon2, lat2 = (radians(coord) for coord in point2[:2])
    dlat = (lat2 - lat1)
    dlon = (lon2 - lon1)
    a = (
        sin(dlat * 0.5)**2 +
        cos(lat1) * cos(lat2) * sin(dlon * 0.5)**2
    )

    return EARTH_MEAN_DIAMETER * asin(sqrt(a))


def _py_distance(point1, point2):
    '''
        Calculating great-circle distance
        (see https://en.wikipedia.org/wiki/Great-circle_distance)
    '''
    lon1, lat1 = (radians(coord) for coord in point1)
    lon2, lat2 = (radians(coord) for coord in point2)

    dlon = fabs(lon1 - lon2)
    dlat = fabs(lat1 - lat2)

    numerator = sqrt(
        (cos(lat2)*sin(dlon))**2 +
        ((cos(lat1)*sin(lat2)) - (sin(lat1)*cos(lat2)*cos(dlon)))**2)

    denominator = (
        (sin(lat1)*sin(lat2)) +
        (cos(lat1)*cos(lat2)*cos(dlon)))

    c = atan2(numerator, denominator)
    return EARTH_MEAN_RADIUS*c


def bearing(point1, point2):
    '''
        Calculating initial bearing between two points
        (see http://www.movable-type.co.uk/scripts/latlong.html)
    '''
    lon1, lat1 = (radians(coord) for coord in point1)
    lon2, lat2 = (radians(coord) for coord in point2)

    dlat = (lat2 - lat1)
    dlon = (lon2 - lon1)
    numerator = sin(dlon) * cos(lat2)
    denominator = (
        cos(lat1) * sin(lat2) -
        (sin(lat1) * cos(lat2) * cos(dlon))
    )

    theta = atan2(numerator, denominator)
    theta_deg = (degrees(theta) + 360) % 360
    return theta_deg


def final_bearing(point1, point2):
    return (bearing(point2, point1) + 180) % 360


def destination(point, distance, bearing):
    '''
        Given a start point, initial bearing, and distance, this will
        calculate the destina­tion point and final bearing travelling
        along a (shortest distance) great circle arc.

        (see http://www.movable-type.co.uk/scripts/latlong.htm)
    '''

    lon1, lat1 = (radians(coord) for coord in point)
    radians_bearing = radians(bearing)

    delta = distance / EARTH_MEAN_RADIUS

    lat2 = asin(
        sin(lat1)*cos(delta) +
        cos(lat1)*sin(delta)*cos(radians_bearing)
    )
    numerator = sin(radians_bearing) * sin(delta) * cos(lat1)
    denominator = cos(delta) - sin(lat1) * sin(lat2)

    lon2 = lon1 + atan2(numerator, denominator)

    lon2_deg = (degrees(lon2) + 540) % 360 - 180
    lat2_deg = degrees(lat2)

    return (lon2_deg, lat2_deg)


def approximate_destination(point, distance, theta):
    # http://stackoverflow.com/questions/2187657/calculate-second-point-knowing-the-starting-point-and-distance
    lon, lat = point
    radians_theta = radians(theta)
    dx = distance*cos(radians_theta)
    dy = distance*sin(radians_theta)

    dlon = dx / (EARTH_EQUATORIAL_METERS_PER_DEGREE*cos(radians(lat)))
    dlat = dy / EARTH_EQUATORIAL_METERS_PER_DEGREE

    return (lon+dlon, lat+dlat)


def _py_from4326_to3857(point):
    '''
        Reproject point from EPSG:4326 to EPSG:3857
        (see
            http://wiki.openstreetmap.org/wiki/Mercator,
            https://epsg.io/4326,
            https://epsg.io/3857)
    '''
    lon, lat = point
    xtile = lon * EARTH_EQUATORIAL_METERS_PER_DEGREE
    ytile = log(tan(radians(45 + lat / 2.0))) * EARTH_EQUATORIAL_RADIUS
    return (xtile, ytile)


def _py_from3857_to4326(point):
    '''
        Reproject point from EPSG:3857 to EPSG:4326
        (see http://wiki.openstreetmap.org/wiki/Mercator)

        Reverse Spherical Mercator:
            λ = E/R + λo
            φ = π/2 - 2*arctan(exp(-N/R))
    '''
    x, y = point
    lon = x / EARTH_EQUATORIAL_METERS_PER_DEGREE
    lat = degrees(2.0 * atan(exp(y/EARTH_EQUATORIAL_RADIUS)) - HALF_PI)
    return (lon, lat)


try:
    from ._sphere import (
        _approximate_distance,
        _haversine_distance,
        _distance,
        _from4326_to3857,
        _from3857_to4326,
    )
    approximate_distance = _approximate_distance
    haversine_distance = _haversine_distance
    distance = _distance
    from4326_to3857 = _from4326_to3857
    from3857_to4326 = _from3857_to4326
except ImportError:  # pragma: no cover
    approximate_distance = _py_approximate_distance
    haversine_distance = _py_haversine_distance
    distance = _py_distance
    from4326_to3857 = _py_from4326_to3857
    from3857_to4326 = _py_from3857_to4326
